@@ -8,21 +8,25 @@ An implementation of a HTTP/2 stream.
 import warnings
 
 from enum import Enum, IntEnum
+from hpack import HeaderTuple
 from hyperframe.frame import (
     HeadersFrame, ContinuationFrame, DataFrame, WindowUpdateFrame,
-    RstStreamFrame, PushPromiseFrame,
+    RstStreamFrame, PushPromiseFrame, AltSvcFrame
 )
 
 from .errors import STREAM_CLOSED
 from .events import (
     RequestReceived, ResponseReceived, DataReceived, WindowUpdated,
     StreamEnded, PushedStreamReceived, StreamReset, TrailersReceived,
-    InformationalResponseReceived,
+    InformationalResponseReceived, AlternativeServiceAvailable
 )
 from .exceptions import (
     ProtocolError, StreamClosedError, InvalidBodyLengthError
 )
-from .utilities import guard_increment_window, is_informational_response
+from .utilities import (
+    guard_increment_window, is_informational_response, authority_from_headers,
+    secure_headers
+)
 
 
 class StreamState(IntEnum):
@@ -51,6 +55,10 @@ class StreamInputs(Enum):
     RECV_CONTINUATION = 12  # Added in 2.0.0
     SEND_INFORMATIONAL_HEADERS = 13  # Added in 2.2.0
     RECV_INFORMATIONAL_HEADERS = 14  # Added in 2.2.0
+    SEND_ALTERNATIVE_SERVICE = 15  # Added in 2.3.0
+    RECV_ALTERNATIVE_SERVICE = 16  # Added in 2.3.0
+    UPGRADE_CLIENT = 17  # Added 2.3.0
+    UPGRADE_SERVER = 18  # Added 2.3.0
 
 
 # This array is initialized once, and is indexed by the stream states above.
@@ -211,6 +219,7 @@ class H2StreamStateMachine(object):
         """
         assert self.client is None
         self.client = False
+        self.headers_received = True
         return []
 
     def recv_new_pushed_stream(self, previous_state):
@@ -221,6 +230,7 @@ class H2StreamStateMachine(object):
         """
         assert self.client is None
         self.client = True
+        self.headers_sent = True
         return []
 
     def send_push_promise(self, previous_state):
@@ -300,15 +310,90 @@ class H2StreamStateMachine(object):
         assert previous_state == StreamState.CLOSED
         raise ProtocolError("Attempted to push on closed stream.")
 
+    def send_informational_response(self, previous_state):
+        """
+        Called when an informational header block is sent (that is, a block
+        where the :status header has a 1XX value).
+
+        Only enforces that these are sent *before* final headers are sent.
+        """
+        if self.headers_sent:
+            raise ProtocolError("Information response after final response")
+
+        return []
+
     def recv_informational_response(self, previous_state):
         """
         Called when an informational header block is received (that is, a block
         where the :status header has a 1XX value).
         """
+        if self.headers_received:
+            raise ProtocolError("Informational response after final response")
+
         event = InformationalResponseReceived()
         event.stream_id = self.stream_id
         return [event]
 
+    def recv_alt_svc(self, previous_state):
+        """
+        Called when receiving an ALTSVC frame.
+
+        RFC 7838 allows us to receive ALTSVC frames at any stream state, which
+        is really absurdly overzealous. For that reason, we want to limit the
+        states in which we can actually receive it. It's really only sensible
+        to receive it after we've sent our own headers and before the server
+        has sent its header block: the server can't guarantee that we have any
+        state around after it completes its header block, and the server
+        doesn't know what origin we're talking about before we've sent ours.
+
+        For that reason, this function applies a few extra checks on both state
+        and some of the little state variables we keep around. If those suggest
+        an unreasonable situation for the ALTSVC frame to have been sent in,
+        we quietly ignore it (as RFC 7838 suggests).
+
+        This function is also *not* always called by the state machine. In some
+        states (IDLE, RESERVED_LOCAL, CLOSED) we don't bother to call it,
+        because we know the frame cannot be valid in that state (IDLE because
+        the server cannot know what origin the stream applies to, CLOSED
+        because the server cannot assume we still have state around,
+        RESERVED_LOCAL because by definition if we're in the RESERVED_LOCAL
+        state then *we* are the server).
+        """
+        # Servers can't receive ALTSVC frames, but RFC 7838 tells us to ignore
+        # them.
+        if self.client is False:
+            return []
+
+        # If we've received the response headers from the server they can't
+        # guarantee we still have any state around. Other implementations
+        # (like nghttp2) ignore ALTSVC in this state, so we will too.
+        if self.headers_received:
+            return []
+
+        # Otherwise, this is a sensible enough frame to have received. Return
+        # the event and let it get populated.
+        return [AlternativeServiceAvailable()]
+
+    def send_alt_svc(self, previous_state):
+        """
+        Called when sending an ALTSVC frame on this stream.
+
+        For consistency with the restrictions we apply on receiving ALTSVC
+        frames in ``recv_alt_svc``, we want to restrict when users can send
+        ALTSVC frames to the situations when we ourselves would accept them.
+
+        That means: when we are a server, when we have received the request
+        headers, and when we have not yet sent our own response headers.
+        """
+        # We should not send ALTSVC after we've sent response headers, as the
+        # client may have disposed of its state.
+        if self.headers_sent:
+            raise ProtocolError(
+                "Cannot send ALTSVC after sending response headers."
+            )
+
+        return
+
 
 # STATE MACHINE
 #
@@ -393,6 +478,13 @@ _transitions = {
     (StreamState.IDLE, StreamInputs.RECV_PUSH_PROMISE):
         (H2StreamStateMachine.recv_new_pushed_stream,
             StreamState.RESERVED_REMOTE),
+    (StreamState.IDLE, StreamInputs.RECV_ALTERNATIVE_SERVICE):
+        (None, StreamState.IDLE),
+    (StreamState.IDLE, StreamInputs.UPGRADE_CLIENT):
+        (H2StreamStateMachine.request_sent, StreamState.HALF_CLOSED_LOCAL),
+    (StreamState.IDLE, StreamInputs.UPGRADE_SERVER):
+        (H2StreamStateMachine.request_received,
+            StreamState.HALF_CLOSED_REMOTE),
 
     # State: reserved local
     (StreamState.RESERVED_LOCAL, StreamInputs.SEND_HEADERS):
@@ -407,6 +499,10 @@ _transitions = {
         (None, StreamState.CLOSED),
     (StreamState.RESERVED_LOCAL, StreamInputs.RECV_RST_STREAM):
         (H2StreamStateMachine.stream_reset, StreamState.CLOSED),
+    (StreamState.RESERVED_LOCAL, StreamInputs.SEND_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.send_alt_svc, StreamState.RESERVED_LOCAL),
+    (StreamState.RESERVED_LOCAL, StreamInputs.RECV_ALTERNATIVE_SERVICE):
+        (None, StreamState.RESERVED_LOCAL),
 
     # State: reserved remote
     (StreamState.RESERVED_REMOTE, StreamInputs.RECV_HEADERS):
@@ -422,6 +518,8 @@ _transitions = {
         (None, StreamState.CLOSED),
     (StreamState.RESERVED_REMOTE, StreamInputs.RECV_RST_STREAM):
         (H2StreamStateMachine.stream_reset, StreamState.CLOSED),
+    (StreamState.RESERVED_REMOTE, StreamInputs.RECV_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.recv_alt_svc, StreamState.RESERVED_REMOTE),
 
     # State: open
     (StreamState.OPEN, StreamInputs.SEND_HEADERS):
@@ -449,9 +547,13 @@ _transitions = {
     (StreamState.OPEN, StreamInputs.RECV_PUSH_PROMISE):
         (H2StreamStateMachine.recv_push_promise, StreamState.OPEN),
     (StreamState.OPEN, StreamInputs.SEND_INFORMATIONAL_HEADERS):
-        (None, StreamState.OPEN),
+        (H2StreamStateMachine.send_informational_response, StreamState.OPEN),
     (StreamState.OPEN, StreamInputs.RECV_INFORMATIONAL_HEADERS):
         (H2StreamStateMachine.recv_informational_response, StreamState.OPEN),
+    (StreamState.OPEN, StreamInputs.SEND_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.send_alt_svc, StreamState.OPEN),
+    (StreamState.OPEN, StreamInputs.RECV_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.recv_alt_svc, StreamState.OPEN),
 
     # State: half-closed remote
     (StreamState.HALF_CLOSED_REMOTE, StreamInputs.SEND_HEADERS):
@@ -479,6 +581,13 @@ _transitions = {
         (H2StreamStateMachine.send_reset, StreamState.CLOSED),
     (StreamState.HALF_CLOSED_REMOTE, StreamInputs.RECV_CONTINUATION):
         (H2StreamStateMachine.send_reset, StreamState.CLOSED),
+    (StreamState.HALF_CLOSED_REMOTE, StreamInputs.SEND_INFORMATIONAL_HEADERS):
+        (H2StreamStateMachine.send_informational_response,
+            StreamState.HALF_CLOSED_REMOTE),
+    (StreamState.HALF_CLOSED_REMOTE, StreamInputs.SEND_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.send_alt_svc, StreamState.HALF_CLOSED_REMOTE),
+    (StreamState.HALF_CLOSED_REMOTE, StreamInputs.RECV_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.recv_alt_svc, StreamState.HALF_CLOSED_REMOTE),
 
     # State: half-closed local
     (StreamState.HALF_CLOSED_LOCAL, StreamInputs.RECV_HEADERS):
@@ -499,12 +608,21 @@ _transitions = {
     (StreamState.HALF_CLOSED_LOCAL, StreamInputs.RECV_PUSH_PROMISE):
         (H2StreamStateMachine.recv_push_promise,
             StreamState.HALF_CLOSED_LOCAL),
+    (StreamState.HALF_CLOSED_LOCAL, StreamInputs.RECV_INFORMATIONAL_HEADERS):
+        (H2StreamStateMachine.recv_informational_response,
+            StreamState.HALF_CLOSED_LOCAL),
+    (StreamState.HALF_CLOSED_LOCAL, StreamInputs.SEND_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.send_alt_svc, StreamState.HALF_CLOSED_LOCAL),
+    (StreamState.HALF_CLOSED_LOCAL, StreamInputs.RECV_ALTERNATIVE_SERVICE):
+        (H2StreamStateMachine.recv_alt_svc, StreamState.HALF_CLOSED_LOCAL),
 
     # State: closed
     (StreamState.CLOSED, StreamInputs.RECV_WINDOW_UPDATE):
         (H2StreamStateMachine.window_updated, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.RECV_RST_STREAM):
         (None, StreamState.CLOSED),  # Swallow further RST_STREAMs
+    (StreamState.CLOSED, StreamInputs.RECV_ALTERNATIVE_SERVICE):
+        (None, StreamState.CLOSED),
 
     # While closed, all other received frames should cause RST_STREAM
     # frames to be emitted. END_STREAM is always carried *by* a frame,
@@ -561,6 +679,9 @@ class H2Stream(object):
         # The actual received content length. Always tracked.
         self._actual_content_length = 0
 
+        # The authority we believe this stream belongs to.
+        self._authority = None
+
     @property
     def open(self):
         """
@@ -581,6 +702,22 @@ class H2Stream(object):
         """
         return self.state_machine.state == StreamState.CLOSED
 
+    def upgrade(self, client_side):
+        """
+        Called by the connection to indicate that this stream is the initial
+        request/response of an upgraded connection. Places the stream into an
+        appropriate state.
+        """
+        assert self.stream_id == 1
+        input_ = (
+            StreamInputs.UPGRADE_CLIENT if client_side
+            else StreamInputs.UPGRADE_SERVER
+        )
+
+        # This may return events, we deliberately don't want them.
+        self.state_machine.process_input(input_)
+        return
+
     def send_headers(self, headers, encoder, end_stream=False):
         """
         Returns a list of HEADERS/CONTINUATION frames to emit as either headers
@@ -631,6 +768,9 @@ class H2Stream(object):
         if self.state_machine.trailers_sent and not end_stream:
             raise ProtocolError("Trailers must have END_STREAM set.")
 
+        if self.state_machine.client and self._authority is None:
+            self._authority = authority_from_headers(headers)
+
         return frames
 
     def push_stream_in_band(self, related_stream_id, headers, encoder):
@@ -695,6 +835,16 @@ class H2Stream(object):
         df.flags.add('END_STREAM')
         return [df]
 
+    def advertise_alternative_service(self, field_value):
+        """
+        Advertise an RFC 7838 alternative service. The semantics of this are
+        better documented in the ``H2Connection`` class.
+        """
+        self.state_machine.process_input(StreamInputs.SEND_ALTERNATIVE_SERVICE)
+        asf = AltSvcFrame(self.stream_id)
+        asf.field = field_value
+        return [asf]
+
     def increase_flow_control_window(self, increment):
         """
         Increase the size of the flow control window for the remote side.
@@ -704,7 +854,10 @@ class H2Stream(object):
         wuf.window_increment = increment
         return [wuf]
 
-    def receive_push_promise_in_band(self, promised_stream_id, headers):
+    def receive_push_promise_in_band(self,
+                                     promised_stream_id,
+                                     headers,
+                                     header_encoding):
         """
         Receives a push promise frame sent on this stream, pushing a remote
         stream. This is called on the stream that has the PUSH_PROMISE sent
@@ -714,10 +867,13 @@ class H2Stream(object):
             StreamInputs.RECV_PUSH_PROMISE
         )
         events[0].pushed_stream_id = promised_stream_id
+
+        if header_encoding:
+            headers = list(_decode_headers(headers, header_encoding))
         events[0].headers = headers
         return [], events
 
-    def remotely_pushed(self):
+    def remotely_pushed(self, pushed_headers):
         """
         Mark this stream as one that was pushed by the remote peer. Must be
         called immediately after initialization. Sends no frames, simply
@@ -726,9 +882,10 @@ class H2Stream(object):
         events = self.state_machine.process_input(
             StreamInputs.RECV_PUSH_PROMISE
         )
+        self._authority = authority_from_headers(pushed_headers)
         return [], events
 
-    def receive_headers(self, headers, end_stream):
+    def receive_headers(self, headers, end_stream, header_encoding):
         """
         Receive a set of headers (or trailers).
         """
@@ -754,6 +911,9 @@ class H2Stream(object):
             if not end_stream:
                 raise ProtocolError("Trailers must have END_STREAM set")
 
+        if header_encoding:
+            headers = list(_decode_headers(headers, header_encoding))
+
         events[0].headers = headers
         return [], events
 
@@ -800,6 +960,29 @@ class H2Stream(object):
         )
         assert False, "Should not be reachable"
 
+    def receive_alt_svc(self, frame):
+        """
+        An Alternative Service frame was received on the stream. This frame
+        inherits the origin associated with this stream.
+        """
+        # If the origin is present, RFC 7838 says we have to ignore it.
+        if frame.origin:
+            return [], []
+
+        events = self.state_machine.process_input(
+            StreamInputs.RECV_ALTERNATIVE_SERVICE
+        )
+
+        # There are lots of situations where we want to ignore the ALTSVC
+        # frame. If we need to pay attention, we'll have an event and should
+        # fill it out.
+        if events:
+            assert isinstance(events[0], AlternativeServiceAvailable)
+            events[0].origin = self._authority
+            events[0].field_value = frame.field
+
+        return [], events
+
     def reset_stream(self, error_code=0):
         """
         Close the stream locally. Reset the stream with an error code.
@@ -829,7 +1012,10 @@ class H2Stream(object):
         """
         Helper method to build headers or push promise frames.
         """
-        headers = ((name.lower(), value) for name, value in headers)
+        # We need to lowercase the header names, and to ensure that secure
+        # header fields are kept out of compression contexts.
+        headers = _lowercase_header_names(headers)
+        headers = secure_headers(headers)
         encoded_headers = encoder.encode(headers)
 
         # Slice into blocks of max_outbound_frame_size. Be careful with this:
@@ -861,7 +1047,7 @@ class H2Stream(object):
         Content-Length header to be present.
         """
         for n, v in headers:
-            if n == 'content-length':
+            if n == b'content-length':
                 try:
                     self._expected_content_length = int(v, 10)
                 except ValueError:
@@ -891,3 +1077,33 @@ class H2Stream(object):
 
             if end_stream and expected != actual:
                 raise InvalidBodyLengthError(expected, actual)
+
+
+def _lowercase_header_names(headers):
+    """
+    Given an iterable of header two-tuples, rebuilds that iterable with the
+    header names lowercased. This generator produces tuples that preserve the
+    original type of the header tuple for tuple and any ``HeaderTuple``.
+    """
+    for header in headers:
+        if isinstance(header, HeaderTuple):
+            yield header.__class__(header[0].lower(), header[1])
+        else:
+            yield (header[0].lower(), header[1])
+
+
+def _decode_headers(headers, encoding):
+    """
+    Given an iterable of header two-tuples and an encoding, decodes those
+    headers using that encoding while preserving the type of the header tuple.
+    This ensures that the use of ``HeaderTuple`` is preserved.
+    """
+    for header in headers:
+        # This function expects to work on decoded headers, which are always
+        # HeaderTuple objects.
+        assert isinstance(header, HeaderTuple)
+
+        name, value = header
+        name = name.decode(encoding)
+        value = value.decode(encoding)
+        yield header.__class__(name, value)
